
# -*- coding: utf-8 -*-
"""
Created on Jun 24 2020

@author: J. C. Vasquez-Correa
"""
import os
import sys

import numpy as np
import pandas as pd
from phonet.phonet import Phonet
from phonet.phonet import Phonological as phon
import scipy.stats as st
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Times New Roman"

PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
from utils import save_dict_kaldimat, get_dict

from script_mananger import script_manager
import torch
from tqdm import tqdm

class Phonological:
    """
    Compute phonological features from continuous speech files.

    18 descriptors are computed, bases on 18 different phonological classes from the phonet toolkit 
    https://phonet.readthedocs.io/en/latest/?badge=latest

    It computes the phonological log-likelihood ratio features from phonet

    Static or dynamic matrices can be computed:

    Static matrix is formed with 108 features formed with (18 descriptors) x (6 functionals: mean, std, skewness, kurtosis, max, min)

    Dynamic matrix is formed with the 18 descriptors computed for frames of 25 ms with a time-shift of 10 ms.


    Script is called as follows

    >>> python phonological.py <file_or_folder_audio> <file_features> <static (true or false)> <plots (true or false)> <format (csv, txt, npy, kaldi, torch)>

    Examples command line:

    >>> python phonological.py "../audios/001_ddk1_PCGITA.wav" "phonologicalfeaturesAst.txt" "true" "true" "txt"
    >>> python phonological.py "../audios/001_ddk1_PCGITA.wav" "phonologicalfeaturesUst.csv" "true" "true" "csv"
    >>> python phonological.py "../audios/001_ddk1_PCGITA.wav" "phonologicalfeaturesUdyn.pt" "false" "true" "torch"

    >>> python phonological.py "../audios/" "phonologicalfeaturesst.txt" "true" "false" "txt"
    >>> python phonological.py "../audios/" "phonologicalfeaturesst.csv" "true" "false" "csv"
    >>> python phonological.py "../audios/" "phonologicalfeaturesdyn.pt" "false" "false" "torch"
    >>> python phonological.py "../audios/" "phonologicalfeaturesdyn.csv" "false" "false" "csv"

    Examples directly in Python

    >>> phonological=Phonological()
    >>> file_audio="../audios/001_ddk1_PCGITA.wav"
    >>> features1=phonological.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
    >>> features2=phonological.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
    >>> features3=phonological.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
    >>> phonological.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")

    """

    def __init__(self):
        phonolist=phon()
        self.head_dyn=phonolist.get_list_phonological_keys()
        self.statistics=["mean", "std", "skewness", "kurtosis", "max", "min"]
        self.head_st=[]
        for j in self.head_dyn:
            for l in self.statistics:
                self.head_st.append(j+"_"+l)
        self.phon=Phonet(["all"])

    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the phonological features from an audio file

        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> phonological=Phonological()
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features1=phonological.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=phonological.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=phonological.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> phonological.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")

        >>> phonological=Phonological()
        >>> path_audio="../audios/"
        >>> features1=phonological.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=phonological.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=phonological.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> phonological.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")

        """
        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")

        if audio.find('.wav') == -1 and audio.find('.WAV') == -1:
            raise ValueError(audio+" is not a valid wav file")
        
        df=self.phon.get_PLLR(audio, plot_flag=plots)

        keys=df.keys().tolist()
        keys.remove('time')

        if static:
            feat=[]
            functions=[np.mean, np.std, st.skew, st.kurtosis, np.max, np.min]
            for j in keys:
                for function in functions:
                    feat.append(function(df[j]))
            feat=np.expand_dims(feat, axis=0)

        else:
            feat=np.stack([df[k] for k in keys], axis=1)

        if fmt in("npy","txt"):
            return feat
        elif fmt in("dataframe","csv") and static:
            dff = {}
            for e, k in enumerate(self.head_st):
                dff[k] = feat[:, e]
            return pd.DataFrame(df)
        elif fmt in("dataframe","csv") and not static:
            return df
        elif fmt=="torch":
            return torch.from_numpy(feat)
        elif fmt=="kaldi":
            featmat=np.stack([df[k] for k in keys], axis=1)
            name_all=audio.split('/')
            dictX={name_all[-1]:featmat}
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")



    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the phonological features for audios inside a path
        
        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> phonological=Phonological()
        >>> path_audio="../audios/"
        >>> features1=phonological.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=phonological.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=phonological.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> phonological.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """

        hf=os.listdir(path_audio)
        hf.sort()

        pbar=tqdm(range(len(hf)))
        ids=[]

        Features=[]
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file=path_audio+hf[j]
            feat=self.extract_features_file(audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))
        
        Features=np.vstack(Features)
        ids=np.hstack(ids)

        return self.save_features(Features, ids, fmt, static, kaldi_file)



    def save_features(self, Features, ids, fmt, static, kaldi_file):

        if static:
            head = self.head_st
        else:
            head = self.head_dyn
    
        if fmt in("npy","txt"):
            return Features
        elif fmt in("dataframe","csv"):
            df = {}
            for e, k in enumerate(head):
                df[k] = Features[:, e]
            df["id"] = ids
            return pd.DataFrame(df)
        elif fmt=="torch":
            return torch.from_numpy(Features)
        elif fmt=="kaldi":
            dictX=get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")


if __name__=="__main__":

    if len(sys.argv)!=6:
        print("python phonological.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>")
        sys.exit()

    phonological=Phonological()
    script_manager(sys.argv, phonological)


import numpy as np
import torch

def script_manager(args, feature_method):

    audio, file_features, static, plots, fmt = check_paramters(args)
    features = extract_features(feature_method, audio, file_features, static, plots, fmt)
    save_features(file_features, fmt, features)

def save_features(file_features, fmt, features):
    if fmt=="npy":
        np.save(file_features, features)
    elif fmt=="txt":
        np.savetxt(file_features, features)
    elif fmt=="csv":
        features.to_csv(file_features, index=False)
    elif fmt=="torch":
        torch.save(features, file_features)

def extract_features(feature_method, audio, file_features, static, plots, fmt):
    features=[]
    if audio.find('.wav')!=-1 or audio.find('.WAV')!=-1:
        if fmt=="kaldi":
            feature_method.extract_features_file(audio, static=static, plots=plots, fmt=fmt, kaldi_file=file_features)
        else:
            features=feature_method.extract_features_file(audio, static=static, plots=plots, fmt=fmt)

    else:
        if fmt=="kaldi":
            feature_method.extract_features_path(audio, static=static, plots=plots, fmt=fmt, kaldi_file=file_features)
        else:
            features=feature_method.extract_features_path(audio, static=static, plots=plots, fmt=fmt)
    return features


def check_paramters(args):
    audio=args[1]
    file_features=args[2]
    if args[3]=="false" or args[3]=="False":
        static=False
    elif args[3]=="true" or args[3]=="True":
        static=True
    else:
        raise ValueError(args[3] +" is not a valid argument for <static>. It should be only True or False")

    if args[4]=="false" or args[4]=="False":
        plots=False
    elif args[4]=="true" or args[4]=="True":
        plots=True
    else:
        raise ValueError(args[4] +" is not a valid argument for <plots>. It should be only True or False")

    if args[5]=="npy" or args[5]=="csv" or args[5]=="txt" or args[5]=="torch" or args[5]=="kaldi":
        fmt=args[5]
    else:
        raise ValueError(args[5]+ " is not a valid argument for <format>. It should be only csv, txt, npy, kaldi, or torch")
    return audio,file_features,static,plots,fmt



# -*- coding: utf-8 -*-
"""
Created on Jul 21 2017

@author: J. C. Vasquez-Correa
"""

from scipy.io.wavfile import read
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Times New Roman"
import matplotlib.mlab as mlab
import pysptk
import pandas as pd
import torch
from tqdm import tqdm
PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
import disvoice.praat.praat_functions as praat_functions
from disvoice.script_mananger import script_manager
from articulation_functions import extract_transitions, get_transition_segments

from utils import dynamic2statict_artic, save_dict_kaldimat, get_dict, fill_when_empty




class Articulation:
    """

    Compute articulation features from continuous speech.

    122 descriptors are computed:

    1-22. Bark band energies in onset transitions (22 BBE).

    23-34. Mel frequency cepstral coefficients in onset transitions (12 MFCC onset)

    35-46. First derivative of the MFCCs in onset transitions (12 DMFCC onset)

    47-58. Second derivative of the MFCCs in onset transitions (12 DDMFCC onset)

    59-80. Bark band energies in offset transitions (22 BBE).

    81-92. MFCCC in offset transitions (12 MFCC offset)

    93-104. First derivative of the MFCCs in offset transitions (12 DMFCC offset)

    105-116. Second derivative of the MFCCs in offset transitions (12 DMFCC offset)

    117. First formant Frequency

    118. First Derivative of the first formant frequency

    119. Second Derivative of the first formant frequency

    120. Second formant Frequency

    121. First derivative of the Second formant Frequency

    122. Second derivative of the Second formant Frequency

    Static or dynamic matrices can be computed:
    
    Static matrix is formed with 488 features formed with (122 descriptors) x (4 functionals: mean, std, skewness, kurtosis)
    
    Dynamic matrix are formed with the 58 descriptors (22 BBEs, 12 MFCC, 12DMFCC, 12 DDMFCC ) computed for frames of 40 ms with a time-shift of 20 ms in onset transitions.

    The first two frames of each recording are not considered for dynamic analysis to be able to stack the derivatives of MFCCs


    Notes:
    1. The first two frames of each recording are not considered for dynamic analysis to be able to stack the derivatives of MFCCs
    2. The fundamental frequency is computed the PRAAT algorithm. To use the RAPT method,  change the "self.pitch method" variable in the class constructor.

    Script is called as follows

    >>> python articulation.py <file_or_folder_audio> <file_features> <static (true or false)> <plots (true or false)> <format (csv, txt, npy, kaldi, torch)>

    Examples command line:

    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKst.txt" "true" "true" txt
    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKst.csv" "true" "true" csv
    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKst.pt" "true" "true" torch
    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKdyn.txt" "false" "true" txt
    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKdyn.csv" "false" "true" csv
    >>> python articulation.py "../audios/001_ddk1_PCGITA.wav" "articulation_featuresDDKdyn.pt" "false" "true" torch
    
    Examples directly in Python

    >>> articulation=Articulation()
    >>> file_audio="../audios/001_ddk1_PCGITA.wav"
    >>> features1=articulation.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
    >>> features2=articulation.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
    >>> features3=articulation.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
    >>> articulation.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")

    """

    def __init__(self):
        self.pitch_method="rapt"
        self.sizeframe=0.04
        self.step=0.02
        self.nB=22
        self.nMFCC=12
        self.minf0=60
        self.maxf0=350
        self.voice_bias=-0.2
        self.len_thr_miliseconds=270.0
        self.head=["BBEon_"+str(j) for j in range(1,23)]
        self.head+=["MFCCon_"+str(j) for j in range(1,13)]
        self.head+=["DMFCCon_"+str(j) for j in range(1,13)]
        self.head+=["DDMFCCon_"+str(j) for j in range(1,13)]
        self.head+=["BBEoff_"+str(j) for j in range(1,23)]
        self.head+=["MFCCoff_"+str(j) for j in range(1,13)]
        self.head+=["DMFCCoff_"+str(j) for j in range(1,13)]
        self.head+=["DDMFCCoff_"+str(j) for j in range(1,13)]
        self.head+=["F1", "DF1", "DDF1", "F2", "DF2", "DDF2"]
        self.head_dyn=["BBEon_"+str(j) for j in range(1,23)]
        self.head_dyn+=["MFCCon_"+str(j) for j in range(1,13)]
        self.head_dyn+=["DMFCCon_"+str(j) for j in range(1,13)]
        self.head_dyn+=["DDMFCCon_"+str(j) for j in range(1,13)]
        self.head_st=[]
        for k in ["avg", "std", "skewness", "kurtosis"]:
            for h in self.head:
                self.head_st.append(k+" "+h)
        if not os.path.exists(PATH+'/../../tempfiles/'):
            os.makedirs(PATH+'/../../tempfiles/')

    def plot_art(self, data_audio,fs,F0,F1,F2,segmentsOn,segmentsOff):
        """Plots of the articulation features

        :param data_audio: speech signal.
        :param fs: sampling frequency
        :param F0: contour of the fundamental frequency
        :param F1: contour of the 1st formant
        :param F2: contour of the 2nd formant
        :param segmentsOn: list with the onset segments
        :param segmentsOff: list with the offset segments
        :returns: plots of the articulation features.
        """
        x_axis='Time (s)'
        y_axis='Frequency (Hz)'
        plt.figure(1)
        plt.subplot(311)
        t=np.arange(0, float(len(data_audio))/fs, 1.0/fs)
        if len(t)>len(data_audio):
            t=t[:len(data_audio)]
        elif len(t)<len(data_audio):
            data_audio=data_audio[:len(t)]
        plt.plot(t, data_audio, 'k')
        plt.ylabel('Amplitude', fontsize=14)
        plt.xlim([0, t[-1]])
        plt.grid(True)

        plt.subplot(312)
        t0=np.linspace(0.0,t[-1],len(F0))
        plt.plot(t0, F0, color='r', linewidth=2.0, label='F0')
        plt.xlabel(x_axis, fontsize=14)
        plt.ylabel(y_axis, fontsize=14)
        plt.ylim([0,np.max(F0)+10])
        plt.xlim([0, t0[-1]])
        plt.grid(True)
        plt.legend()

        plt.subplot(313)
        fsp=int(len(F1)/t[-1])
        t2=np.arange(0.0, t[-1], 1.0/fsp)
        if len(t2)>len(F1):
            t2=t2[:len(F1)]
        elif len(F1)>len(t2):
            F1=F1[:len(t2)]
            F2=F2[:len(t2)]

        plt.plot(t2, F1, color='k', linewidth=2.0, label='F1')
        plt.plot(t2, F2, color='g', linewidth=2.0, label='F2')
        plt.xlabel(x_axis, fontsize=14)
        plt.ylabel(y_axis, fontsize=14)
        plt.ylim([0,np.max(F2)+10])
        plt.xlim([0, t2[-1]])
        plt.grid(True)
        plt.legend()
        plt.tight_layout()    
        plt.show()
        plt.close()
        F0on=[]
        F0off=[]
        for j in range(2, len(F0)):
            if F0[j-1]==0 and F0[j]!=0:
                F0on.append(F0[j-8:j+7])
            elif F0[j-1]!=0 and F0[j]==0:
                F0off.append(F0[j-8:j+7])

        for j in range(1,len(segmentsOn)-1):
            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [1,3]}, sharex=True, figsize=(5,4))
            t=np.arange(len(segmentsOn[j]))/fs
            a0.plot(t[0:640], segmentsOn[j][0:640], color='k', label="unvoiced")
            a0.plot(t[640:], segmentsOn[j][640:], color='k', alpha=0.5, label="voiced")
            a0.grid()
            a0.legend(loc=2, ncol=2 )
            a0.set_xlim([0,0.08])
            spec, freqs, t, im=plt.specgram(segmentsOn[j], NFFT=128, Fs=fs, window=mlab.window_hanning, noverlap=100, detrend=mlab.detrend_mean)
            a1.imshow(np.log10(np.abs(np.flipud(spec))), extent=[0, .08, 1, 4000], aspect='auto', cmap=plt.cm.viridis,
                    vmax=np.log10(np.abs(spec).max()), vmin=np.log10(np.abs(spec).min()), interpolation="bilinear")
            a1.set_ylabel(y_axis, fontsize=12)
            a1.set_xlabel(x_axis, fontsize=12)
            a1.set_xlim([0,0.08])
            a02 = a1.twinx()
            fsp=len(F0on[j])/0.086
            t2=np.arange(len(F0on[j]))/fsp
            a02.plot(t2, F0on[j], 'r', linewidth=2,label="F0")
            a02.set_ylabel(r'$F_0$ (Hz)', color='r', fontsize=12)
            a02.tick_params('y', colors='r')
            a02.set_xlim([0,0.08])
            plt.tight_layout()
            plt.show()

        for j in range(1,len(segmentsOff)-1):
            f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [1,3]}, sharex=True, figsize=(5,4))
            t=np.arange(len(segmentsOff[j]))/fs
            a0.plot(t[0:640], segmentsOff[j][0:640], color='k', label="voiced")
            a0.plot(t[640:], segmentsOff[j][640:], color='k', alpha=0.5, label="unvoiced")
            a0.grid()
            a0.legend(loc=1, ncol=2 )
            
            spec, freqs, t, im=plt.specgram(segmentsOff[j], NFFT=128, Fs=fs, window=mlab.window_hanning, noverlap=100, detrend=mlab.detrend_mean)
            a1.imshow(np.log10(np.abs(np.flipud(spec))), extent=[0, .08, 1, 4000], aspect='auto', cmap=plt.cm.viridis,
                    vmax=np.log10(np.abs(spec).max()), vmin=np.log10(np.abs(spec).min()), interpolation="bilinear")
            a1.set_ylabel(y_axis, fontsize=12)
            a1.set_xlabel(x_axis, fontsize=12)
            a02 = a1.twinx()
            fsp=len(F0off[j])/0.086
            t2=np.arange(len(F0off[j]))/fsp
            a02.plot(t2, F0off[j], 'r', linewidth=2,label="F0")
            a02.set_ylabel(r'$F_0$ (Hz)', color='r', fontsize=12)
            a02.tick_params('y', colors='r')
            plt.tight_layout()
            plt.show()


    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the articulation features from an audio file

        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> articulation=Articulation()
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features1=articulation.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=articulation.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=articulation.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> articulation.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")
        
        >>> path_audio="../audios/"
        >>> features1=articulation.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=articulation.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=articulation.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> articulation.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")

        """
        fs, data_audio=read(audio)

        if len(data_audio.shape)>1:
            data_audio = data_audio.mean(1)
        data_audio=data_audio-np.mean(data_audio)
        data_audio=data_audio/float(np.max(np.abs(data_audio)))
        size_frameS=self.sizeframe*float(fs)
        size_stepS=self.step*float(fs)

        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")

        F0, segmentsOn, segmentsOff = self.extract_transition_segments(audio, fs, data_audio, size_stepS)

        BBEon, MFCCon=extract_transitions(segmentsOn, fs, size_frameS, size_stepS, self.nB, self.nMFCC)
        BBEoff, MFCCoff=extract_transitions(segmentsOff, fs, size_frameS, size_stepS, self.nB, self.nMFCC)

        DMFCCon=np.asarray([np.diff(MFCCon[:,nf], n=1) for nf in range(MFCCon.shape[1])]).T
        DDMFCCon=np.asarray([np.diff(MFCCon[:,nf], n=2) for nf in range(MFCCon.shape[1])]).T

        DMFCCoff=np.asarray([np.diff(MFCCoff[:,nf], n=1) for nf in range(MFCCoff.shape[1])]).T
        DDMFCCoff=np.asarray([np.diff(MFCCoff[:,nf], n=2) for nf in range(MFCCoff.shape[1])]).T


        name_audio=audio.split('/')
        temp_uuid='artic'+name_audio[-1][0:-4]

        temp_filename=PATH+'/../../tempfiles/tempFormants'+temp_uuid+'.txt'
        praat_functions.praat_formants(audio, temp_filename,self.sizeframe,self.step)
        [F1, F2]=praat_functions.decodeFormants(temp_filename)
        os.remove(temp_filename)

        if len(F0)<len(F1):
            F0=np.hstack((F0, np.zeros(len(F1)-len(F0))))
            F1nz=np.zeros((0,1))
            F2nz=np.zeros((0,1))
            DF1=np.zeros((0,1))
            DDF1=np.zeros((0,1))
            DF2=np.zeros((0,1))
            DDF2=np.zeros((0,1))

        else:
            F1=np.hstack((F1, np.zeros(len(F0)-len(F1))))
            F2=np.hstack((F2, np.zeros(len(F0)-len(F2))))

            pos0=np.where(F0==0)[0]
            dpos0=np.hstack(([1],np.diff(pos0)))
            f0u=np.split(pos0, np.where(dpos0>1)[0])

            thr_sil=int(self.len_thr_miliseconds/self.step)

            len_segments=np.array([len(segment) for segment in f0u])
            index_silence=np.where(len_segments>=thr_sil)[0]
            F1[index_silence]=0
            F2[index_silence]=0

            F1nz=F1[F1!=0]
            F2nz=F2[F2!=0]
            DF1=np.diff(F1, n=1)
            DF2=np.diff(F2, n=1)
            DDF1=np.diff(F1, n=2)
            DDF2=np.diff(F2, n=2)


            if plots:
                self.plot_art(data_audio,fs,F0,F1,F2,segmentsOn,segmentsOff)

            F1nz=fill_when_empty(F1nz)
            F2nz=fill_when_empty(F2nz)
            DF1=fill_when_empty(DF1)
            DDF1=fill_when_empty(DDF1)
            DF2=fill_when_empty(DF2)
            DDF2=fill_when_empty(DDF2)

        if static:
            feat_mat=dynamic2statict_artic([BBEon, MFCCon, DMFCCon, DDMFCCon, BBEoff, MFCCoff, DMFCCoff, DDMFCCoff, F1nz, DF1, DDF1, F2nz, DF2, DDF2])
            feat_mat=np.expand_dims(feat_mat,0)
            head=self.head_st
        else:
            feat_mat=np.hstack((BBEon[2:,:], MFCCon[2:,:], DMFCCon[1:,:], DDMFCCon))
            head=self.head_dyn

        if fmt in("npy","txt"):
            return feat_mat
        elif fmt in("dataframe","csv"):
            df={}
            for e, k in enumerate(head):
                df[k]=feat_mat[:,e]
            return pd.DataFrame(df)
        elif fmt=="torch":
            feat_t=torch.from_numpy(feat_mat)
            return feat_t
        elif fmt=="kaldi":
            name_all=audio.split('/')
            dictX={name_all[-1]:feat_mat}
            save_dict_kaldimat(dictX, kaldi_file)

    def extract_transition_segments(self, audio, fs, data_audio, size_stepS):
        if self.pitch_method == 'praat':
            name_audio=audio.split('/')
            temp_uuid='articulation'+name_audio[-1][0:-4]
            temp_filename_vuv=PATH+'/../../tempfiles/tempVUV'+temp_uuid+'.txt'
            temp_filename_f0=PATH+'/../../tempfiles/tempF0'+temp_uuid+'.txt'
            praat_functions.praat_vuv(audio, temp_filename_f0, temp_filename_vuv, time_stepF0=self.step, minf0=self.minf0, maxf0=self.maxf0)
            F0,_=praat_functions.decodeF0(temp_filename_f0,len(data_audio)/float(fs),self.step)
            segmentsFull,segmentsOn,segmentsOff=praat_functions.read_textgrid_trans(temp_filename_vuv,data_audio,fs,self.sizeframe)
            os.remove(temp_filename_vuv)
            os.remove(temp_filename_f0)
        elif self.pitch_method == 'rapt':
            data_audiof=np.asarray(data_audio*(2**15), dtype=np.float32)
            F0=pysptk.sptk.rapt(data_audiof, fs, int(size_stepS), min=self.minf0, max=self.maxf0, voice_bias=self.voice_bias, otype='f0')

            segmentsOn=get_transition_segments(F0, data_audio, fs, 'onset')
            segmentsOff=get_transition_segments(F0, data_audio, fs, 'offset')
        return F0,segmentsOn,segmentsOff


    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the articulation features for audios inside a path
        
        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> articulation=Articulation()
        >>> path_audio="../audios/"
        >>> features1=articulation.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=articulation.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=articulation.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> articulation.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """

        hf=os.listdir(path_audio)
        hf.sort()

        pbar=tqdm(range(len(hf)))
        ids=[]

        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")        

        Features=[]
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file=path_audio+hf[j]
            feat=self.extract_features_file(audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))
        
        Features=np.vstack(Features)
        ids=np.hstack(ids)
        if fmt in("npy","txt"):
            return Features
        if fmt in("dataframe","csv"):
            df={}
            if static:
                head=self.head_st
            else:
                head=self.head_dyn
            for e, k in enumerate(head):
                df[k]=Features[:,e]

            df["id"]=ids
            return pd.DataFrame(df)
        if fmt=="torch":
            return torch.from_numpy(Features)
        if fmt=="kaldi":
            dictX=get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)


if __name__=="__main__":

    if len(sys.argv)!=6:
        print("python articulation.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>")
        sys.exit()

    articulation=Articulation()
    script_manager(sys.argv, articulation)



import math
import numpy as np
import pysptk

def bark(f):
    x=(f*0.00076)
    x2=(f/7500)**2
    b=[]
    for i in range (0,len(f)):
        b.append(13*( math.atan(x[i]) )+3.5*( math.atan(x2[i]))) #Bark scale values
    return (b)

def barke(x,Fs, nfft=2048, nB=25):
    """
    e: Energy in frequency bands according to the Bark scale
    x: signal
    Fs: sampling frequency
    nfft: number of points for the Fourier transform
    nB: number of bands for energy computation
    """
    eps = 1e-30
    y = fftsolp(x,nfft)
    f = (Fs/2)*(np.linspace(0,1,int(nfft/2+1)))
    barkScale = bark(f)
    barkIndices = []
    for i in range (0,len(barkScale)):
        barkIndices.append(int(barkScale[i]))

    barkIndices = np.asarray(barkIndices)

    barkEnergy=[]
    for i in range (nB):
        brk = np.nonzero(barkIndices==i)
        brk = np.asarray(brk)[0]
        sizeb=len(brk)
        if (sizeb>0):
            barkEnergy.append(sum(np.abs(y[brk]))/sizeb)
        else:
            barkEnergy.append(0)

    e = np.asarray(barkEnergy)+eps
    e = np.log(e)
    return e


def fftsolp(x,nfft):
    """
    STFT for compute the energy in Bark scale
    x: signal
    nffft: number of points of the Fourier transform
    """
    window = np.hamming(len(x)/4)
    noverlap = np.ceil(len(window)/2)

    nx = len(x)
    nwind = len(window)

    ncol = np.fix((nx-noverlap)/(nwind-noverlap))
    ncol = int(ncol)
    colindex = (np.arange(0,ncol))*(nwind-noverlap)
    colindex = colindex.astype(int)

    rowindex = np.arange(0,nwind)
    rowindex = rowindex.astype(int)
    rowindex = rowindex[np.newaxis]
    rowindex = rowindex.T
    d = np.ones((nwind,ncol),dtype=np.int)
    y = x[d*(rowindex+colindex)]
    window = window.astype(float)
    window = window[np.newaxis]
    window = window.T
    new = window*d
    y = new*y
    y = y[:,0]

    y = np.fft.fft(y,nfft)
    y = (y[0:int(nfft/2+1)])
    return y

    
def extract_transitions(segments, fs, size_frameS, size_stepS, nB=22, nMFCC=12, nfft=2048):
    frames=[]
    size_frame_full=int(2**np.ceil(np.log2(size_frameS)))
    fill=int(size_frame_full-size_frameS)
    overlap=size_stepS/size_frameS
    for j in range(len(segments)):
        if (len(segments[j])>size_frameS):
            nF=int((len(segments[j])/size_frameS)/overlap)-1
            for iF in range(nF):
                frames.append(np.hamming(size_frameS)*segments[j][int(iF*size_stepS):int(iF*size_stepS+size_frameS)])

    BarkEn=np.zeros((len(frames),nB))
    MFCC=np.zeros((len(frames),nMFCC))
    for j in range(len(frames)):
        frame_act=np.hstack((frames[j], np.zeros(fill)))
        BarkEn[j,:]=barke(frame_act,fs, nfft, nB)
        MFCC[j,:]=pysptk.sptk.mfcc(frame_act, order=nMFCC, fs=fs, alpha=0.97, num_filterbanks=32, cepslift=22, use_hamming=True)
    return BarkEn, MFCC


def get_transition_segments(F0, data_audio, fs, transition, size_tran=0.04):
    segment=[]
    time_stepF0=int(len(data_audio)/len(F0))

    for j in range(2, len(F0)):
        if transition=="onset":
            condition=F0[j-1]==0 and F0[j]!=0
        elif transition=="offset":
            condition=F0[j-1]!=0 and F0[j]==0
        if condition:
            border=j*time_stepF0
            initframe=int(border-size_tran*fs)
            endframe=int(border+size_tran*fs)
            segment.append(data_audio[initframe:endframe])
    return segment




# -*- coding: utf-8 -*-
"""
Feature extraction from speech signals based on representation learning strategies
"""

import os
import sys
from scipy.io.wavfile import read
import torch
from librosa.feature import melspectrogram
import numpy as np
import warnings

import matplotlib.pyplot as plt
import pandas as pd
import scipy.stats as st

PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
from CAE import CAEn
from RAE import RAEn
class AEspeech:

    def __init__(self, model, units):
        """
        Feature extraction from speech signals based on representation learning strategies using convolutional and recurrent autoencoders
        
        :param model: type of autoencoder to extract the features from ('CAE': convolutional autoencoders, 'RAE': recurrent autoencoder)
        :param units: number of hidden neurons in the bottleneck space (64, 128, 256, 512, 1024)
        :returns: AEspeech Object.
        
        """
        self.model_type=model
        self.units=units
        self.min_scaler=-50.527256
        self.max_scaler=6.8561997
        if model=="CAE":
            self.AE=CAEn(units)
            if torch.cuda.is_available():
                
                self.AE.load_state_dict(torch.load(PATH+"/"+str(units)+'_CAE.pt'))
                self.AE.cuda()
            else:
                self.AE.load_state_dict(torch.load(PATH+"/"+str(units)+'_CAE.pt', map_location='cpu'))
        elif model=="RAE":
            self.AE=RAEn(units)
            if torch.cuda.is_available():
                self.AE.load_state_dict(torch.load(PATH+"/"+str(units)+'_RAE.pt'))
                self.AE.cuda()
            else:
                self.AE.load_state_dict(torch.load(PATH+"/"+str(units)+'_RAE.pt', map_location='cpu'))

        else:
            raise ValueError("Model "+model+" is not valid. Please choose only CAE or RAE")

        self.feat_names_bottle_all, self.feat_names_error_all = self.get_feature_names()
    
    def compute_spectrograms(self, wav_file):
        """
        Compute the tensor of Mel-scale spectrograms to be used as input for the autoencoders from a wav file
        
        :param wav_file: .wav file with a sampling frequency of 16kHz
        :returns: Pytorch tensor (N, C, F, T). N: batch of spectrograms extracted every 500ms, C: number of channels (1),  F: number of Mel frequencies (128), T: time steps (126)
        
        """
        NFFT=512
        FRAME_SIZE=0.5
        TIME_SHIFT=0.25
        HOP=64
        NMELS=128

        if wav_file.find('.wav')==-1 and wav_file.find('.WAV')==-1:
            raise ValueError(wav_file+" is not a valid audio file")
        fs, signal=read(wav_file)

        if len(signal.shape)>1:
            signal = signal.mean(1)
        if fs!=16000:
            raise ValueError(str(fs)+" is not a valid sampling frequency")
        signal=signal-np.mean(signal)
        signal=signal/np.max(np.abs(signal))
        init=0
        endi=int(FRAME_SIZE*fs)
        nf=int(len(signal)/(TIME_SHIFT*fs))-1
        if nf>0:
            mat=torch.zeros(nf,1, NMELS,126)
            j=0
            for k in range(nf):
                try:
                    frame=signal[init:endi]
                    imag=melspectrogram(y=frame, sr=fs, n_fft=NFFT, hop_length=HOP, n_mels=NMELS, fmax=fs/2)
                    init=init+int(TIME_SHIFT*fs)
                    endi=endi+int(TIME_SHIFT*fs)
                    if np.min(np.min(imag))<=0:
                        warnings.warn("There is Inf values in the Mel spectrogram")
                        continue
                    imag=np.log(imag, dtype=np.float32)
                    imagt=torch.from_numpy(imag)
                    mat[j,:,:,:]=imagt
                    j+=1
                except:
                    init=init+int(TIME_SHIFT*fs)
                    endi=endi+int(TIME_SHIFT*fs)
                    warnings.warn("There is non valid values in the wav file")
        else:
            raise ValueError("WAV file is too short to compute the Mel spectrogram tensor")
        
        return mat[0:j,:,:,:]


    def show_spectrograms(self, spectrograms):
        """
        Visualization of the computed tensor of Mel-scale spectrograms to be used as input for the autoencoders from a wav file
        
        :param spectrograms: tensor of spectrograms obtained from ``compute_spectrograms(wav-file)``
        
        """
        mmax=2595*np.log10(1+8000/700)
        m=np.linspace(0,mmax,11)

        f=np.round(700*(10**(m/2595)-1))
        f=f[::-1]
        for k in range(spectrograms.shape[0]):
            fig,  ax=plt.subplots(1, 1)
            fig.set_size_inches(5, 5)
            mat=spectrograms[k,:,:]#.data.numpy()[k,0,:,:]
            ax.imshow(np.flipud(mat), cmap=plt.cm.viridis, vmax=mat.max())
            ax.set_yticks(np.linspace(0,128,11))
            ax.set_yticklabels(map(str, f))
            ax.set_xticks(np.linspace(0,126,6))
            ax.set_xticklabels(map(str, np.linspace(0,500,6, dtype=np.int)))
            ax.set_ylabel("Frequency (Hz)")
            ax.set_xlabel("Time (ms)")
            plt.tight_layout()
            plt.show()


    def standard(self, tensor):
        """
        standardize input tensor for the autoencoders

        :param tensor: input tensor for the AEs (N, 128,126)
        :returns:  standardize tensor for the AEs (N, 128,126)
        
        """
        temp=tensor-self.min_scaler
        temp=temp/(self.max_scaler-self.min_scaler)
        return temp.float()

    def destandard(self, tensor):
        """
        destandardize input tensor from the autoencoders
        
        :param tensor: standardized input tensor for the AEs (N, 128,126)
        :returns:  destandardized tensor for the AEs (N, 128,126)
        
        """
        temp=tensor*(self.max_scaler-self.min_scaler)
        return temp+self.min_scaler

    def compute_bottleneck_features(self, wav_file, return_numpy=True):
        """
        Compute the the bottleneck features of the autoencoder
        
        :param wav_file: .wav file with a sampling frequency of 16kHz
        :param return_numpy: return the features in a numpy array (True) or a Pytorch tensor (False)
        :returns: Pytorch tensor (nf, h) or numpy array (nf, h) with the extracted features. nf: number of frames, size of the bottleneck space
        
        """

        mat=self.compute_spectrograms(wav_file)
        mat=self.standard(mat)
        if torch.cuda.is_available():
            mat=mat.cuda()
        to, bot=self.AE.forward(mat)
        if return_numpy:
            return bot.data.numpy()
        return bot

    def compute_rec_error_features(self, wav_file, return_numpy=True):
        """
        Compute the  reconstruction error features from the autoencoder
        
        :param wav_file: .wav file with a sampling frequency of 16kHz
        :param return_numpy: return the features in a numpy array (True) or a Pytorch tensor (False)
        :returns: Pytorch tensor (nf, 128) or numpy array (nf, 128) with the extracted features. nf: number of frames
        
        """
        mat=self.compute_spectrograms(wav_file)
        mat=self.standard(mat)
        if torch.cuda.is_available():
            mat=mat.cuda()
        to, bot=self.AE.forward(mat)
        
        mat_error=(mat[:,0,:,:]-to[:,0,:,:])**2
        to=self.destandard(to)
        error=torch.mean(mat_error,2)
        if return_numpy:
            return error.data.numpy()
        return error



    def compute_rec_spectrogram(self, wav_file, return_numpy=True):
        """
        Compute the  reconstructed spectrogram from the autoencoder
        
        :param wav_file: .wav file with a sampling frequency of 16kHz
        :param return_numpy: return the features in a numpy array (True) or a Pytorch tensor (False)
        :returns: Pytorch tensor (N, C, F, T). N: batch of spectrograms extracted every 500ms, C: number of channels (1),  F: number of Mel frequencies (128), T: time steps (126)
        
        """
        mat=self.compute_spectrograms(wav_file)
        mat=self.standard(mat)
        if torch.cuda.is_available():
            mat=mat.cuda()
        to, bot=self.AE.forward(mat)        
        to=self.destandard(to)

        if return_numpy:
            return to.data.numpy(), mat.data.numpy()
        return to, mat


    def plot_spectrograms(self, wav_file):
        """
        Figure of the decoded spectrograms by the AEs
        
        :param wav_file: .wav file with a sampling frequency of 16kHz
        
        """
        decmat, encmat =self.compute_rec_spectrogram(wav_file)

        index=np.arange(0,decmat.shape[0], 2)
        decall=decmat[index,0,:,:]
        encall=encmat[index,0,:,:]
        decspec=np.concatenate(decall,axis=1)
        encspec=np.concatenate(encall,axis=1)

        mmax=2595*np.log10(1+8000/700)
        m=np.linspace(0,mmax,6)

        f=np.round(700*(10**(m/2595)-1))
        f=f[::-1]
        fig,  (ax1, ax2)=plt.subplots(2, 1, sharex=True)
        fig.set_size_inches(8, 4)
        ax1.imshow(np.flipud(encspec), cmap=plt.cm.viridis, vmax=encspec.max())
        ax1.set_yticks(np.linspace(0,128,6))
        ax1.set_yticklabels(map(str, f))
        ax1.set_xticks(np.linspace(0,encspec.shape[1],6))
        ax1.set_xticklabels(map(str, np.linspace(0,0.5*len(index),6, dtype=np.int)))
        ax1.set_ylabel("Frequency (Hz)")
        ax1.set_title("Input spectrogram")
        ax2.imshow(np.flipud(decspec), cmap=plt.cm.viridis, vmax=decspec.max())
        ax2.set_yticks(np.linspace(0,128,6))
        ax2.set_yticklabels(map(str, f))
        ax2.set_xticks(np.linspace(0,encspec.shape[1],6))
        ax2.set_xticklabels(map(str, np.linspace(0,0.5*len(index),6, dtype=np.int)))
        ax2.set_ylabel("Frequency (Hz)")
        ax2.set_xlabel("Time (s)")
        ax2.set_title("Decoded spectrogram")
        plt.tight_layout()
        plt.show()

        plt.figure(figsize=(5,4))
        m=np.linspace(0,mmax,128)

        f=np.round(700*(10**(m/2595)-1))
        f=f[::-1]
        error_all=np.abs(encspec/encspec.max()-decspec/decspec.max())**2
        avg_er=np.mean(error_all, axis=1)
        std_er=np.std(error_all, axis=1)
        plt.plot(f,avg_er, 'k')
        plt.fill_between(f,avg_er-std_er, avg_er+std_er, color='k', alpha=0.2)
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Mean squared reconstruction error")
        plt.grid()
        plt.tight_layout()
        plt.show()

        
    def compute_dynamic_features(self, wav_directory):
        """
        Compute both the bottleneck and the reconstruction error features from the autoencoder for wav files inside a directory
        
        :param wav_directory: .wav file with a sampling frequency of 16kHz
        :return: dictionary with the extracted bottleneck and error features, and with information about which frame coresponds to which wav file in the directory.
        
        """

        if os.path.isdir(wav_directory):
            hf=os.listdir(wav_directory)
            hf.sort()
        else:
            raise ValueError(wav_directory+" is not a valid directory")

        if wav_directory[-1]!='/':
            wav_directory=wav_directory+"/"


        metadata={"wav_file":[], "frame": [], "bottleneck": [], "error":[]}
        for wav_file in hf:
            bottle=self.compute_bottleneck_features(wav_directory+wav_file, True)
            error=self.compute_rec_error_features(wav_directory+wav_file, True)
            metadata["bottleneck"].append(bottle)
            metadata["error"].append(error)
            nframes=error.shape[0]
            list_wav=np.repeat(wav_file, nframes)
            metadata["wav_file"].append(list_wav)
            frames=np.arange(nframes)
            metadata["frame"].append(frames)

        metadata["bottleneck"]=np.concatenate(metadata["bottleneck"], 0)
        metadata["error"]=np.concatenate(metadata["error"], 0)
        metadata["wav_file"]=np.hstack(metadata["wav_file"])
        metadata["frame"]=np.hstack(metadata["frame"])
        return metadata


    def compute_global_features(self, wav_directory, stack_feat=False):
        """
        Compute global features (1 vector per utterance) both for the bottleneck and the reconstruction error features from the autoencoder for wav files inside a directory 
        
        :param wav_directory: .wav file with a sampling frequency of 16kHz
        :param stack_feat: if True, returns also a feature matrix with the stack of the bottleneck and error features
        :return: pandas dataframes with the bottleneck and error features.
        
        """

        if os.path.isdir(wav_directory):
            hf=os.listdir(wav_directory)
            hf.sort()
        else:
            raise ValueError(wav_directory+" is not a valid directory")

        if wav_directory[-1]!='/':
            wav_directory=wav_directory+"/"

        

        if stack_feat:
            feat_names_all=self.feat_names_bottle_all+self.feat_names_error_all

        bottle_feat=np.zeros((len(hf), len(self.feat_names_bottle_all)))
        error_feat=np.zeros((len(hf), len(self.feat_names_error_all)))
        
        if stack_feat:
            feat_all=np.zeros((len(hf),len(self.feat_names_bottle_all)+len(self.feat_names_error_all) ))

        for i, wav_file in enumerate(hf):
            try:
                bottle=self.compute_bottleneck_features(wav_directory+wav_file, True)
                bottle_feat[i,:]=np.hstack((np.mean(bottle, 0), np.std(bottle, 0), st.skew(bottle, 0), st.kurtosis(bottle, 0)))
                error=self.compute_rec_error_features(wav_directory+wav_file, True)
                error_feat[i,:]=np.hstack((np.mean(error, 0), np.std(error, 0), st.skew(error, 0), st.kurtosis(error, 0)))
            except:
                warnings.warn("ERROR WITH "+wav_file)
                continue

        dict_feat_bottle={}
        dict_feat_bottle["ID"]=hf
        for j in range(bottle_feat.shape[1]):
            dict_feat_bottle[self.feat_names_bottle_all[j]]=bottle_feat[:,j]

        dict_feat_error={}
        dict_feat_error["ID"]=hf
        for j in range(error_feat.shape[1]):
            dict_feat_error[self.feat_names_error_all[j]]=error_feat[:,j]

        df1=pd.DataFrame(dict_feat_bottle)
        df2=pd.DataFrame(dict_feat_error)

        if stack_feat:
            feat_all=np.concatenate((bottle_feat, error_feat), axis=1)

            dict_feat_all={}
            dict_feat_all["ID"]=hf
            for j in range(feat_all.shape[1]):
                dict_feat_all[feat_names_all[j]]=feat_all[:,j] 

            df3=pd.DataFrame(dict_feat_all)

            return df1, df2, df3

        return df1, df2

    def get_feature_names(self):
        feat_names_bottle=["bottleneck_"+str(k) for k in range(self.units)]
        feat_names_error=["error_"+str(k) for k in range(128)]

        stat_names=["avg", "std", "skewness", "kurtosis"]

        feat_names_bottle_all=[]
        feat_names_error_all=[]

        for k in stat_names:
            for j in feat_names_bottle:
                feat_names_bottle_all.append(k+"_"+j)
            for j in feat_names_error:
                feat_names_error_all.append(k+"_"+j)
        return feat_names_bottle_all,feat_names_error_all





import os
import numpy as np
import sys
PATH = os.path.dirname(os.path.realpath(__file__))

sys.path.append(PATH)

def multi_find(s, r):
	"""
	Internal function used to decode the Formants file generated by Praat.
	"""
	s_len = len(s)
	r_len = len(r)
	_complete = []
	for i in range(s_len):
		# search for r in s until not enough characters are left
		if s[i:i + r_len] == r:
			_complete.append(i)
		else:
			i = i + 1
	return _complete

def praat_vuv(audio_filaname, resultsp, resultst, time_stepF0=0, minf0=75, maxf0=600, maxVUVPeriod=0.02, averageVUVPeriod=0.01):
	"""
	runs vuv_praat script to obtain pitch and voicing decisions for a wav file.
	It writes the results into two text files, one for the pitch and another
	for the voicing decisions. These results can then be read using the function
	read_textgrid_trans and decodeF0

	:param audio_filaname: Full path to the wav file
	:param resultsp: Full path to the resulting file with the pitch
	:param resultst: Full path to the resulting file with the voiced/unvoiced decisions
	:param time_stepF0: time step to compute the pitch, default value is 0 and Praat will use 0.75 / minf0
	:param minf0: minimum frequency for the pitch in Hz, default is 75Hz
	:param maxf0: maximum frequency for the pitch in Hz, default is 600
	:param maxVUVPeriod: maximum interval that considered part of a larger voiced interval, default 0.02
	:param averageVUVPeriod: half of this value will be taken to be the amount to which a voiced interval will extend 							beyond its initial and final points, default is 0.01
	:returns: nothing
	"""
	if sys.platform.find('win')>=0:
		command='/praat.exe --run '+PATH+'/vuv_praat.praat '
	else:
		command='praat '+PATH+'/vuv_praat.praat '
	command+=audio_filaname+' '+resultsp +' '+  resultst+' '
	command+=str(minf0)+' '+str(maxf0)+' '
	command+=str(time_stepF0)+' '+str(maxVUVPeriod)+' '+str(averageVUVPeriod)
	os.system(command)

def praat_formants(audio_filename, results_filename,sizeframe,step, n_formants=5, max_formant=5500):
	"""
	runs FormantsPraat script to obtain the formants for a wav file.
	It writes the results into a text file.
	These results can then be read using the function decodeFormants.

	:param audio_filaname: Full path to the wav file, string
	:param results_filename: Full path to the resulting file with the formants
	:param sizeframe: window size 
	:param step: time step to compute formants
	:param n_formants: number of formants to look for
	:param max_formant: maximum frequencyof formants to look for
	:returns: nothing
	"""

	if sys.platform.find('win')>=0:
		command='praat.exe --run '+PATH+'/FormantsPraat.praat '
	else:
		command='praat '+PATH+'/FormantsPraat.praat '
	command+=audio_filename + ' '+results_filename+' '
	command+=str(n_formants)+' '+ str(max_formant) + ' '
	command+=str(float(sizeframe)/2)+' '
	command+=str(float(step))
	os.system(command) #formant extraction praat

def read_textgrid_trans(file_textgrid, data_audio, fs, win_trans=0.04):
	"""
	This function reads a text file with the text grid with voiced/unvoiced
	decisions then finds the onsets (unvoiced -> voiced) and
	offsets (voiced -> unvoiced) and then reads the audio data to returns
	lists of segments of lenght win_trans around these transitions.

	:param file_textgrid: The text file with the text grid with voicing decisions.
	:param data_audio: the audio signal.
	:param fs: sampling frequency of the audio signal.
	:param win_trans: the transition window lenght, default 0.04
	:returns segments: List with both onset and offset transition segments.
	:returns segments_onset: List with onset transition segments
	:returns segments_offset: List with offset transition segments
	"""
	segments=[]
	segments_onset=[]
	segments_offset=[]
	prev_trans=""
	prev_line=0
	with open(file_textgrid) as fp:
		for line in fp:
			line = line.strip('\n')
			if line in ('"V"', '"U"'):
				transVal=int(float(prev_line)*fs)-1
				segment=data_audio[int(transVal-win_trans*fs):int(transVal+win_trans*fs)]
				segments.append(segment)
				if prev_trans in ('"V"', ""):
					segments_onset.append(segment)
				elif prev_trans=='"U"':
					segments_offset.append(segment)
				prev_trans=line
			prev_line=line
	return segments,segments_onset,segments_offset

def decodeF0(fileTxt,len_signal=0, time_stepF0=0):
	"""
	Reads the content of a pitch file created with praat_vuv function.
	By default it will return the contents of the file in two arrays,
	one for the actual values of pitch and the other with the time stamps.
	Optionally the lenght of the signal and the time step of the pitch
	values can be provided to return an array with the full pitch contour
	for the signal, with padded zeros for unvoiced segments.

	:param fileTxt: File with the pitch, which can be generated using the function praat_vuv
	:param len_signal: Lenght of the audio signal in
	:param time_stepF0: The time step of pitch values. Optional.
	:returns pitch: Numpy array with the values of the pitch.
	:returns time_voiced: time stamp for each pitch value.
	"""
	if os.stat(fileTxt).st_size==0:
		return np.array([0]), np.array([0])
	pitch_data=np.loadtxt(fileTxt)
	if len(pitch_data.shape)>1:
		time_voiced=pitch_data[:,0] # First column is the time stamp vector
		pitch=pitch_data[:,1] # Second column
	elif len(pitch_data.shape)==1: # Only one point of data
		time_voiced=pitch_data[0] # First datum is the time stamp
		pitch=pitch_data[1] # Second datum is the pitch value
	if len_signal>0:
		n_frames=int(len_signal/time_stepF0)
		t=np.linspace(0.0,len_signal,n_frames)
		pitch_zeros=np.zeros(int(n_frames))
		if len(pitch_data.shape)>1:
			for idx,time_p in enumerate(time_voiced):
				argmin=np.argmin(np.abs(t-time_p))
				pitch_zeros[argmin]=pitch[idx]
		else:
			argmin=np.argmin(np.abs(t-time_voiced))
			pitch_zeros[argmin]=pitch
		return pitch_zeros, t
	return pitch, time_voiced

def decodeFormants(fileTxt):
	"""
	Read the praat textgrid file for formants and return the array
	
	:param fileTxt: File with the formants, which can be generated using the '''praat_formants'''
	:returns F1: Numpy array containing the values for the first formant
	:returns F2: Numpy array containing the values for the second formant
	"""
	with open(fileTxt, 'r') as fid:
		datam=fid.read()
	fid.close()
	end_line1=multi_find(datam, '\n')
	F1=[]
	F2=[]
	ji=10
	while (ji<len(end_line1)-1):
		line1=datam[end_line1[ji]+1:end_line1[ji+1]]
		cond=(line1 in ('3', '4', '5'))
		if (cond):
			F1.append(float(datam[end_line1[ji+1]+1:end_line1[ji+2]]))
			F2.append(float(datam[end_line1[ji+3]+1:end_line1[ji+4]]))
		ji=ji+1
	F1=np.asarray(F1)
	F2=np.asarray(F2)
	return F1, F2




# -*- coding: utf-8 -*-
"""
Created on Jul 21 2017

@author: J. C. Vasquez-Correa
"""

from scipy.io.wavfile import read
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Times New Roman"

import pysptk

import pandas as pd
PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
from phonation_functions import jitter_env, get_log_energy, shimmer_env, APQ, PPQ

from utils import dynamic2statict, save_dict_kaldimat,get_dict
import praat.praat_functions as praat_functions
from script_mananger import script_manager
import torch
from tqdm import tqdm


class Phonation:
    """
    Compute phonation features from sustained vowels and continuous speech.

    For continuous speech, the features are computed over voiced segments

    Seven descriptors are computed:

    1. First derivative of the fundamental Frequency

    2. Second derivative of the fundamental Frequency

    3. Jitter

    4. Shimmer

    5. Amplitude perturbation quotient

    6. Pitch perturbation quotient

    7. Logaritmic Energy

    Static or dynamic matrices can be computed:

    Static matrix is formed with 29 features formed with (seven descriptors) x (4 functionals: mean, std, skewness, kurtosis) + degree of Unvoiced

    Dynamic matrix is formed with the seven descriptors computed for frames of 40 ms.

    Notes:

    1. In dynamic features the first 11 frames of each recording are not considered to be able to stack the APQ and PPQ descriptors with the remaining ones.
    2. The fundamental frequency is computed the RAPT algorithm. To use the PRAAT method,  change the "self.pitch method" variable in the class constructor.

    Script is called as follows

    >>> python phonation.py <file_or_folder_audio> <file_features> <static (true or false)> <plots (true or false)> <format (csv, txt, npy, kaldi, torch)>

    Examples command line:

    >>> python phonation.py "../audios/001_a1_PCGITA.wav" "phonationfeaturesAst.txt" "true" "true" "txt"
    >>> python phonation.py "../audios/098_u1_PCGITA.wav" "phonationfeaturesUst.csv" "true" "true" "csv"
    >>> python phonation.py "../audios/098_u1_PCGITA.wav" "phonationfeaturesUdyn.pt" "false" "true" "torch"

    >>> python phonation.py "../audios/" "phonationfeaturesst.txt" "true" "false" "txt"
    >>> python phonation.py "../audios/" "phonationfeaturesst.csv" "true" "false" "csv"
    >>> python phonation.py "../audios/" "phonationfeaturesdyn.pt" "false" "false" "torch"

    Examples directly in Python

    >>> from disvoice.phonation import Phonation
    >>> phonation=Phonation()
    >>> file_audio="../audios/001_a1_PCGITA.wav"
    >>> features=phonation.extract_features_file(file_audio, static, plots=True, fmt="numpy")
    >>> features2=phonation.extract_features_file(file_audio, static, plots=True, fmt="dataframe")
    >>> features3=phonation.extract_features_file(file_audio, dynamic, plots=True, fmt="torch")
    
    >>> path_audios="../audios/"
    >>> features1=phonation.extract_features_path(path_audios, static, plots=False, fmt="numpy")
    >>> features2=phonation.extract_features_path(path_audios, static, plots=False, fmt="torch")
    >>> features3=phonation.extract_features_path(path_audios, static, plots=False, fmt="dataframe")

    """
    def __init__(self):
        self.pitch_method="rapt"
        self.size_frame=0.04
        self.size_step=0.02
        self.minf0=60
        self.maxf0=350
        self.voice_bias=-0.2
        self.energy_thr_percent=0.025
        self.head_dyn=["DF0", "DDF0", "Jitter", "Shimmer", "apq", "ppq", "logE"]
        
        if not os.path.exists(PATH+'/../tempfiles/'):
            os.makedirs(PATH+'/../tempfiles/')
        self.head_st=[]
        for k in ["avg", "std", "skewness", "kurtosis"]:
            for h in self.head_dyn:
                self.head_st.append(k+" "+h)


    def plot_phon(self, data_audio,fs,F0,logE):
        """Plots of the phonation features

        :param data_audio: speech signal.
        :param fs: sampling frequency
        :param F0: contour of the fundamental frequency
        :param logE: contour of the log-energy
        :returns: plots of the phonation features.
        """
        plt.figure(figsize=(6,6))
        plt.subplot(211)
        ax1=plt.gca()
        t=np.arange(len(data_audio))/float(fs)
        ax1.plot(t, data_audio, 'k', label="speech signal", alpha=0.8)
        ax1.set_ylabel('Amplitude', fontsize=12)
        ax1.set_xlabel('Time (s)', fontsize=12)
        ax1.set_xlim([0, t[-1]])
        plt.grid(True)
        ax2 = ax1.twinx()
        fsp=len(F0)/t[-1]
        t2=np.arange(len(F0))/fsp
        ax2.plot(t2, F0, 'r', linewidth=2,label=r"F_0")
        ax2.set_ylabel(r'$F_0$ (Hz)', color='r', fontsize=12)
        ax2.tick_params('y', colors='r')

        plt.grid(True)

        plt.subplot(212)
        Esp=len(logE)/t[-1]
        t2=np.arange(len(logE))/float(Esp)
        plt.plot(t2, logE, color='k', linewidth=2.0)
        plt.xlabel('Time (s)', fontsize=14)
        plt.ylabel('Energy (dB)', fontsize=14)
        plt.xlim([0, t[-1]])
        plt.grid(True)
        plt.tight_layout()
        plt.show()

    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the phonation features from an audio file
        
        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> phonation=Phonation()
        >>> file_audio="../audios/001_a1_PCGITA.wav"
        >>> features1=phonation.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=phonation.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=phonation.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> phonation.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")
        """
        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")

        if audio.find('.wav') == -1 and audio.find('.WAV') == -1:
            raise ValueError(audio+" is not a valid wav file")

        fs, data_audio=read(audio)
        data_audio=data_audio-np.mean(data_audio)
        data_audio=data_audio/float(np.max(np.abs(data_audio)))
        size_frameS=self.size_frame*float(fs)
        size_stepS=self.size_step*float(fs)
        overlap=size_stepS/size_frameS
        F0 = self.get_F0(audio, fs, data_audio, size_stepS)
        F0nz=F0[F0!=0]
        Jitter=jitter_env(F0nz, len(F0nz))
        nF=int((len(data_audio)/size_frameS/overlap))-1
        Amp=[]
        logE=[]
        apq=[]
        ppq=[]
        DF0=np.diff(F0nz, 1)
        DDF0=np.diff(DF0,1)
        lnz=0
        for l in range(nF):
            if F0[l]==0:
                continue
            data_frame=data_audio[int(l*size_stepS):int(l*size_stepS+size_frameS)]
            energy=10*get_log_energy(data_frame)
            Amp.append(np.max(np.abs(data_frame)))
            logE.append(energy)
            if lnz>=12:
                amp_arr=np.asarray([Amp[j] for j in range(lnz-12, lnz)])
                apq.append(APQ(amp_arr))
            if lnz>=6:
                f0arr=np.asarray([F0nz[j] for j in range(lnz-6, lnz)])
                ppq.append(PPQ(1./f0arr))
            lnz=lnz+1
        
        Shimmer=shimmer_env(Amp, len(Amp))
        apq=np.asarray(apq)
        ppq=np.asarray(ppq)
        logE=np.asarray(logE)

        if len(apq)==0:
            apq=Shimmer

        if plots:
            self.plot_phon(data_audio,fs,F0,logE)

        if static:
            feat=dynamic2statict([DF0, DDF0, Jitter, Shimmer, apq, ppq, logE])
            feat=np.expand_dims(feat, axis=0)
            head = self.head_st
        else:
            if len(Shimmer)==len(apq):
                feat=np.vstack((DF0[5:], DDF0[4:], Jitter[6:], Shimmer[6:], apq[6:], ppq, logE[6:])).T
            else:
                feat=np.vstack((DF0[11:], DDF0[10:], Jitter[12:], Shimmer[12:], apq, ppq[6:], logE[12:])).T
            head=self.head_dyn

        if fmt in("npy","txt"):
            return feat

        elif fmt in("dataframe","csv"):
            df = {}
            for e, k in enumerate(head):
                df[k] = feat[:, e]
            return pd.DataFrame(df)
        elif fmt=="torch":
            return torch.from_numpy(feat)
        elif fmt=="kaldi":
            name_all=audio.split('/')
            dictX={name_all[-1]:feat}
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")

    def get_F0(self, audio, fs, data_audio, size_stepS):
        if self.pitch_method == 'praat':
            name_audio=audio.split('/')
            temp_uuid='phon'+name_audio[-1][0:-4]
            temp_filename_vuv=PATH+'/../tempfiles/tempVUV'+temp_uuid+'.txt'
            temp_filename_f0=PATH+'/../tempfiles/tempF0'+temp_uuid+'.txt'
            praat_functions.praat_vuv(audio, temp_filename_f0, temp_filename_vuv, time_stepF0=self.size_step, minf0=self.minf0, maxf0=self.maxf0)
            F0,_=praat_functions.decodeF0(temp_filename_f0,len(data_audio)/float(fs),self.size_step)
            os.remove(temp_filename_vuv)
            os.remove(temp_filename_f0)
        elif self.pitch_method == 'rapt':
            data_audiof=np.asarray(data_audio*(2**15), dtype=np.float32)
            F0=pysptk.sptk.rapt(data_audiof, fs, int(size_stepS), min=self.minf0, max=self.maxf0, voice_bias=self.voice_bias, otype='f0')
        return F0

    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the phonation features for audios inside a path
        
        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> phonation=Phonation()
        >>> path_audio="../audios/"
        >>> features1=phonation.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=phonation.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=phonation.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> phonation.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """
        hf=os.listdir(path_audio)
        hf.sort()

        pbar=tqdm(range(len(hf)))
        ids=[]

        Features=[]
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file=path_audio+hf[j]
            feat=self.extract_features_file(audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))
        
        Features=np.vstack(Features)
        ids=np.hstack(ids)
        return self.save_features(Features, ids, fmt, static, kaldi_file)



    def save_features(self, Features, ids, fmt, static, kaldi_file):
        if static:
            head = self.head_st
        else:
            head = self.head_dyn
        
        if fmt in ("npy", "txt"):
            return Features
        if fmt in ("dataframe", "csv"):
            df = {}
            for e, k in enumerate(head):
                df[k] = Features[:, e]
            df["id"] = ids
            return pd.DataFrame(df)
        if fmt == "torch":
            return torch.from_numpy(Features)
        if fmt == "kaldi":
            dictX = get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)

if __name__=="__main__":

    if len(sys.argv)!=6:
        print("python phonation.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>")
        sys.exit()

    phonation=Phonation()
    script_manager(sys.argv, phonation)


try:
    from setuptools import setup, find_packages #enables develop
except ImportError:
    from distutils.core import setup, find_packages
import pathlib

install_requires = [
                    'kaldi_io',
                    'tqdm',
                    'matplotlib',
                    'numpy',
                    'torch',
                    'librosa',
                    'pandas',
                    'pysptk',
                    'phonet',
                    'scipy',
                    'scikit_learn',
                    ]

HERE = pathlib.Path(__file__).parent
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

setup(
    name='disvoice',
    version='0.1.6',
    packages=find_packages(),
    description='Python framework designed to compute different types of features from speech files',
    author='J. C. Vasquez-Correa',
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author_email='juan.vasquez@fau.de',
    url='https://github.com/jcvasquezc/disvoice',
    download_url='https://github.com/jcvasquezc/disvoice/#files',
    project_urls={
        "Source code": "https://github.com/jcvasquezc/disvoice",
        "Documentation": "https://disvoice.readthedocs.io/en/latest/",
        "Bug Tracker": "https://github.com/jcvasquezc/DisVoice/issues"
    },
    license='MIT',
    install_requires=install_requires,
    package_data={'': ['audios/*', "*.praat"]},
    keywords = ['speech', 'speech features', 'articulatory features', 'phoneme recognition', 'prosody', 'praat'],
    dependency_links=['git+git://github.com/jameslyons/python_speech_features'],
    classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",

  ],

)




      








# -*- coding: utf-8 -*-


# -*- coding: utf-8 -*-

from .CAE import CAEn
from .RAE import RAEn
from .AEspeech import AEspeech
from .replearning import RepLearning

__all__=['RepLearning', 'AEspeech', 'CAEn', 'RAEn']